import java.util.*;
import java.net.*;
import java.io.*;

import com.microstar.xml.*;

import gstd.*;

/**
 * Dataset - parses and stores a gstd dataset
 *           from an xml file and stores it in memory
 *           in a form that can easily be visualized
 *
 * @author <a href='mailto:mallett@cs.ualberta.ca'>Daniel Mallett</a>
 */
public class Dataset implements XmlHandler, Constants {

    public Dataset(URL url, float[] snapShot, int numObjects) throws IOException {
	this.snapShot = snapShot;
	this.numObjects = numObjects;

	if (url != null) {
	    readData(url);
	}
    }

    /**
     * parse the XML data file into a data structure to be used for animation
     *
     * @author Daniel Mallett (mallett@cs.ualberta.ca)
     */
    public Vector[] readData(URL url) throws IOException {
	this.url = url;
	XmlParser parser = new XmlParser();
	parser.setHandler(this);
	
	//initialize animation data structure
	data = new Vector[snapShot.length];
	for (int i = 0; i < snapShot.length; i++) {
	    data[i] = new Vector(numObjects);
	}

	//parse the xml
	try {
	    parser.parse(url.toString(),null,url.openStream(),null);
	} catch (IOException e) {
	    error("IO Exception: " + e.getMessage());
	} catch (Exception e) {
	    error("Parsing Error: " + e.getMessage());
	}
	/*
	for (int i = 0; i < data.length; i++) {
	    for (int j = 0; j < data[i].size(); j++) {
		float x[] = (float[])data[i].elementAt(j);
		System.out.print("data[" + i + "].vecData[" + j + "] = " + ((float[])data[i].elementAt(j))[0] + " " + ((float[])data[i].elementAt(j))[1]);
		if (x.length == 4) {
		    System.out.print(" " + ((float[])data[i].elementAt(j))[2] + " " + ((float[])data[i].elementAt(j))[3]);
		}
		System.out.println();
	    }
	}
	*/
	return data;
    }

    /**
     * get object type
     */
    public char getObjectType() {
	return chrObjectType;
    }

    /**
     * return array of data
     */
    public Vector[] getData() {
	return data;
    }

    /**
     * return a name for this dataset
     */
    public String getName() {
	return url.toString();
    }

    /**
     * return the initial position of the object with the given id
     *
     * @author Daniel Mallett (mallett@cs.ualberta.ca)
     */
    private float[] getInitialExtent(int id) {
	return (float[])(data[0].elementAt(id - firstID));
    }


    // XML event handlers
    //
    public void startDocument() { }
    public void endDocument() { }
    public void startElement(String name) { 
	//System.out.println(name + " start");

	if(name.equalsIgnoreCase("instance")) {
	    //only update extent if time stamp is valid
	    if (lastValid) {
		lastExtent = new float[extent.length];
		for (int i = 0; i < extent.length; i++) {
		    lastExtent[i] = extent[i];
		    //System.out.println("lastExtent[" + i + "] = " + lastExtent[i]);
		}
	    }
	    extent = new float[extent.length];
	    extentIndex = 0;

	    return;
	} 


	if (name.equalsIgnoreCase("gstd")) {
	    //initialize the extents
	    short extentSize = NDIMS;
	    if (chrObjectType == RECTANGLE_OBJECT_TYPE) {
		extentSize = NDIMS * 2;
	    }
	    extent = new float[extentSize];
	    for (int i =0; i < extent.length; i++) {
		extent[i] = 0;
	    }
	    extentIndex = 0;
	    return;
	}

	if (name.equalsIgnoreCase("x") ||
	    name.equalsIgnoreCase("y")) {
	    return;
	}

    }

    public void endElement(String name) {	
	//System.out.println(name + " end");

	if (name.equalsIgnoreCase("gstd")) {
	    return;
	}

	if (name.equalsIgnoreCase("instance")) {
	    //System.out.println("id = " + id);
	    //System.out.println("ts = " + ts + " snapShotIndex = " + snapShotIndex + " oldTS = " + lastTS);
	    //System.out.println("extent[0] = " + extent[0] + " extent[1] = " + extent[1]);
	    //System.out.println("lastExtent[o] = " + lastExtent[0] + " lastExtent[1] = " + lastExtent[1]);
	    if (ts == 0.0) {
		/* initialization phase */
		//ignore extra initialization instances beyond numObects
		if (data[snapShotIndex].size() < numObjects) {
		    //System.out.println("aaa trying to add to data snapShotIndex = " + snapShotIndex + " extent = " + extent);
		    data[snapShotIndex].addElement(extent);
		} 
		return;
	    } else if (lastTS == 0.0) {
		//initialization phase over when lastTS == 0 & ts != 0.0
		snapShotIndex = 1;
	    }

	    // data load
	    //System.out.println("objectCount = " + objectCount + " numObjects = " + numObjects);
	    if (objectCount <= numObjects) {
		if (lastID != id) {
		    //new object
		    objectCount++;
		    snapShotIndex = 1;
		    lastTS = 0;
		    lastExtent = getInitialExtent(id);
		}

		while (ts > snapShot[snapShotIndex]) {
		    //System.out.println("bbb trying to add to data snapShotIndex = " + snapShotIndex + " lastExtent = " + lastExtent[0] + " " + lastExtent[1]);
		    data[snapShotIndex++].addElement(lastExtent);
		}
		//System.out.println("after loop ts = " + ts + " valid = " + valid);
		if (ts == snapShot[snapShotIndex]) {
		    if (valid) {
			//System.out.println("ccc trying to add to data snapShotIndex = " + snapShotIndex + " extent = " + extent[0] + " " + extent[1]);
			data[snapShotIndex++].addElement(extent);
		    } else {
			//System.out.println("ddd trying to add to data snapShotIndex = " + snapShotIndex + " lastExtent = " + lastExtent[0] + " " + lastExtent[1]);
			data[snapShotIndex++].addElement(lastExtent);
		    }
		}
	    }
	    return;
	}

	if (valid) {
	    if (name.equalsIgnoreCase("x") || name.equalsIgnoreCase("y")) {
		extent[extentIndex++] = new Float(currentCharData).floatValue();
	    }
	}
    }

    public void ignorableWhitespace(char ch[], int start, int length) { }
    public void charData(char ch[], int start, int length) { 
	currentCharData = new String(ch, start, length);
    }
    public void attribute(String name, String value, boolean isSpecified) { 
	//System.out.println("attrib name = " + name + " value = " + value );
	if(name.equalsIgnoreCase("type")) {
	    if (value.charAt(0) == POINT_OBJECT_TYPE || 
		value.charAt(0) == RECTANGLE_OBJECT_TYPE) {
		chrObjectType = value.charAt(0);
	    } else {
		error("Invalid object type");
	    }
	} else if (name.equalsIgnoreCase("valid")) {
	    lastValid = valid;
	    if (value.equals("f") || value.equals("false")) {
		valid = false;
	    } else {
		valid = true;
	    }
	} else if (name.equalsIgnoreCase("time")) {
	    lastTS = ts;

	    try {
		ts = new Float(value).floatValue();
	    } catch (Exception e) {
		error("Time is invalid: " + e.getMessage());
	    }
	} else if (name.equalsIgnoreCase("id")) {
	    lastID = id;
	    try {
		id = new Integer(value).intValue();
	    } catch (Exception e) {
		error("Id is invalid: " + e.getMessage());
	    }
	    if (firstID == -1) {
		firstID = id;
	    }
	}
    }

    public Object resolveEntity(String publicId, String systemId) { return null; }
    public void startExternalEntity(String systemId) { }
    public void endExternalEntity(String systemId) { }
    public void doctypeDecl(String name, String pubid, String sysid) { }
    public void processingInstruction(String target, String data) { }
    public void error(String message, String url, int line, int column) { 
	System.out.println("Applet error. " + message + " at " + url.toString() + ": line " + line + " column " + column);
    }
 
    private void error(String message) {
	error(message,url.toString(),0,0);
    }

    // location of the xml file to animate
    private URL url;

    // xml input data variables
    private String currentCharData = null;
    private char chrObjectType;  //type of objects: points or rectangles
    private boolean valid = true;
    private int id;
    private float ts = (float)0.0;
    private float extent[];
    private int extentIndex = 0;

    // vars for building animation data structure
    private int numObjects;
    private int objectCount = 0;
    private int snapShotIndex = 0;
    private float snapShot[];
    private int lastID;
    private float lastExtent[];
    private float lastTS = (float)0.0;
    private int firstID = -1;
    private boolean lastValid = true;

    // structure to hold animation data
    private Vector data[];
}
