/**
 * Runs the GSTD as a standalone application that takes command line parameters
 *
 * @author <a href='mailto:mallett@cs.ualberta.ca'>Daniel Mallett</a>
 * @version %I%, %G%
 */
package gstd;
import java.util.*;

public class GSTD implements Constants {
    public static void main(String args[]) {
	if (args.length < NUM_ARGUMENTS) {
	    usage();
	    System.exit(1);
	}
	//do some error checking on the parameters
	
	//create a GSTD instance and set the GSTD parameters
	//only use the XML version if the file neds with .xml
	GSTD_Abstract gstd;
	if (args[i].endsWith(".xml")) {
	    gstd = new GSTD_XML(args[i++]);
	} else if (args[i].endsWith(".csv")) {
	    gstd = new GSTD_CSV(args[i++]);
	} else {
	    gstd = new GSTD_SpaceDelimited(args[i++]);
	}
	gstd.setNumObjects(new Integer(args[i++]).intValue());
	gstd.setStartingID(new Integer(args[i++]).intValue());
	gstd.setDensity(stod(args[i++]));
	gstd.setNumSnapShots(new Integer(args[i++]).intValue());
	gstd.setSeed(new Integer(args[i++]).intValue());
	gstd.setObjectType(new Character(args[i++].charAt(0)).charValue());

	//the distribution patterns (ie. g_0.5_0,5)
	String strInitialDist = args[i++];
	String strTimeDist = args[i++];
	String strDIntervalDist = args[i++];
	String strCenterDist = args[i++];
	String strExtentDist=  args[i++];

	//the approach
	gstd.setApproach(new Character(args[i++].charAt(0)).charValue());

	//extent arrays
	float arrInitialMin[] = new float[NDIMS];
	float arrInitialMax[] = new float[NDIMS];
	for (int i = 0; i < NDIMS; i++) {
	    arrInitialMin[i] = 0;
	    arrInitialMax[i] = 1;
	}
	float arrMinTime[] = new float[] {stod(args[i++])};
	float arrMaxTime[] = new float[] {stod(args[i++])};
	float arrMinDInterval[] = new float[] {stod(args[i++])};
	float arrMaxDInterval[] = new float[] {stod(args[i++])};
	float arrMinCenter[] = setExtent(args);
	float arrMaxCenter[] = setExtent(args);
	float arrMinExtent[] = setExtent(args);
	float arrMaxExtent[] = setExtent(args);

	//create and set the distribution objects

	gstd.setInitialDistribution(getDistribution(strInitialDist,arrInitialMin,arrInitialMax));
	gstd.setTimeDistribution(getDistribution(strTimeDist,arrMinTime,arrMaxTime));
	gstd.setDIntervalDistribution(getDistribution(strDIntervalDist,arrMinDInterval,arrMaxDInterval));
	gstd.setCenterDistribution(getDistribution(strCenterDist,arrMinCenter,arrMaxCenter));
	gstd.setExtentDistribution(getDistribution(strExtentDist,arrMinExtent,arrMaxExtent));

	//handle optional framework generation parameters
	if (args.length == NUM_ARGUMENTS + NUM_OPTIONAL_ARGUMENTS) {
	    //create and use a framework - XXX only works for 2D
	    int   intFrameworkNumRectangles = (new Integer(args[i++]).intValue());
	    float fltFrameworkMaxSizeX = stod(args[i++]);
	    float fltFrameworkMaxSizeY = stod(args[i++]);
	    char  chrFrameworkExtentType = (new Character(args[i++].charAt(0)).charValue());
	    float fltFrameworkDensity = FRAMEWORK_DENSITY;

	    //XXX only XML format frameworks supported at this time
	    GSTD_Abstract frameworkGSTD = new GSTD_XML("framework.xml");
	    frameworkGSTD.setNumObjects(intFrameworkNumRectangles);
	    frameworkGSTD.setStartingID(1);
	    frameworkGSTD.setDensity(fltFrameworkDensity);
	    frameworkGSTD.setNumSnapShots(1);
	    frameworkGSTD.setSeed(gstd.getSeed());
	    frameworkGSTD.setObjectType(RECTANGLE_OBJECT_TYPE);
	    frameworkGSTD.setApproach(ADJUST_APPROACH);
	    
	    float arrFrameworkMinCenter[] = new float[] { (float)(-fltFrameworkMaxSizeX/2.0),
							  (float)(-fltFrameworkMaxSizeY/2.0)};
	    float arrFrameworkMaxCenter[] = new float[] { (float)(fltFrameworkMaxSizeX/2.0),
							  (float)(fltFrameworkMaxSizeY/2.0)};

	    frameworkGSTD.setInitialDistribution(new DistributionRandom(arrInitialMin,arrInitialMax));
	    frameworkGSTD.setTimeDistribution(new DistributionRandom(arrMinTime,arrMaxTime));
	    frameworkGSTD.setDIntervalDistribution(new DistributionRandom(new float[] {0},
									  new float[] {0}));
	    frameworkGSTD.setCenterDistribution(new DistributionRandom(arrFrameworkMinCenter,
								       arrFrameworkMaxCenter));
	    frameworkGSTD.setExtentDistribution(new DistributionRandom(new float[] {0,0},
								       new float[] {0,0}));

	    System.out.println("Generating Framework...");
	    frameworkGSTD.setFramework(new Framework(intFrameworkNumRectangles));
	    frameworkGSTD.generateSpatioTemporalData(true,false);
	    System.out.println("Done Generating framework into file framework.xml");
	    gstd.setFramework(frameworkGSTD.getFramework());
	    gstd.generateSpatioTemporalData(false,true);
	    return;
	}

	//call GSTD normally
	gstd.generateSpatioTemporalData(false,false);
    }

    /**
     * retrieve the extents from the command line
     *
     * @param args is the command line
     */
    private static float[] setExtent(String[] args) {
	float[] fltArr = new float[NDIMS];
	for (int j = 0; j < NDIMS; j++) {
	    fltArr[j] = stod(args[i++]);
	}
	return fltArr;
    }
    
    /**
     * return the appropriate distribution object based on the command line input
     *
     * @param distPattern contains the distribution information
     * <ul><li>g_x_y is a gaussian with mean x and sigma y
     * <li>s_x is a skew with skew factor x
     * <li>r_x is a random with ?
     * </ul>
     *
     * @param min 
     * @param max 
     */
    public static Distribution getDistribution (String distPattern,
						float[] min, float[] max) {
	//tokenize the distPattern string for "_"
	StringTokenizer st = new StringTokenizer(distPattern,"_");

	//find the right distribution to create
	//*** NOTE NUM_PARAMS -1 is because we 
	//use up the first token checking the distribution type
	switch (st.nextToken().charAt(0)) {
	case GAUSSIAN_DISTRIBUTION:
	    if (st.countTokens() != GAUSSIAN_NUM_PARAMS-1) {
		System.out.println("Error!  Invalid gaussian distribution specified");
		break;
	    } 
	    return new 
		DistributionGaussian(stod(st.nextToken()),
				     stod(st.nextToken()),
				     min,max);

	case SKEWED_DISTRIBUTION:
	    if (st.countTokens() != SKEWED_NUM_PARAMS-1) {
		System.out.println("Error!  Invalid skewed distribution specified");
		break;
	    } 
	    return new 
		DistributionSkewed(stod(st.nextToken()),
				   min,max);

	case RANDOM_DISTRIBUTION:
	    if (st.countTokens() != RANDOM_NUM_PARAMS-1) {
		System.out.println("Error!  Invalid random distribution specified");
		break;
	    }
	    return new DistributionRandom(min,max);
	case USER_DISTRIBUTION:
	    //insert code here to return any distribution you would like 
	    //that conforms to the DistributionInterface
	}
	System.out.println("Error!  \""+ distPattern + "\" is an invalid distribution format");
	System.exit(1);
	return null;
    }

  public static void usage() {
      System.out.println( "ERROR! usage is ...");
      System.out.println( "gstd <filename> <#objects> <starting_id> <density D> <#ss> <seed>\n     <objtype> <distr_init> <distr_t> <distr_dt> <distr_c> <distr_ext> <store>\n     <min_t> <max_t> <min_dt> <max_dt> <min_c[]> <max_c[]> <min_ext[]> <max_ext[]>");
      System.out.println( "filename    : the filename to be generated");
      System.out.println( "#objects    : the cardinality");
      System.out.println( "starting_id : the first id of the objects");
      System.out.println( "density D   : the density of the objects");
      System.out.println( "#ss         : the number of snapshots");
      System.out.println( "seed        : the seed for the random function");
      System.out.println( "objtype     : the type of the objects: point or rectangles");
      System.out.println( "distr_init  : the initial distribution");
      System.out.println( "distr_t     : the distribution of the interval time");
      System.out.println( "distr_dt    : the distribution of the dinterval time");
      System.out.println( "distr_c     : the distribution of the center");
      System.out.println( "distr_ext   : the distribution of the extent of the rectangles");
      System.out.println( "store       : the adjustment or radar approach");
      System.out.println( "min_t       : the min value of the interval time");
      System.out.println( "max_t       : the max value of the interval time");
      System.out.println( "min_dt      : the min value of the dinterval time");
      System.out.println( "max_dt      : the max value of the dinterval time");
      System.out.println( "min_c[]     : the min value of the center point");
      System.out.println( "max_c[]     : the max value of the center point");
      System.out.println( "min_ext[]   : the min value of the rectangle extent");
      System.out.println( "max_ext[]   : the max value of the rectangle extent");
      
      System.out.println( "\nIf the distribution is gaussian (g), it has to be the form:");
      System.out.println( "g_mean_variance, where mean and variance are float numbers");
      
      System.out.println( "\nIf the distribution is skewed (s), it has to be the form:");
      System.out.println( "s_p, where p is the skewedness of the distribution");

      System.out.println( "\nIf the distribution is random (r), it has to be the form:");
      System.out.println( "r");

      System.out.println("\nOptional Framework Parameters");
      System.out.println("# rectangles : the number of rectangles in the framework");
      System.out.println("max_x        : maximum size of the x dimension of each rectangle");
      System.out.println("max_y        : maximum size of the y dimension of each rectangle");
      System.out.println("extent_type  : 'v' for variable or 'c' for constant <- note that constant is not yet supported");
      System.out.println("\nIn order to generate and use a framework all 4 parameters must be present.");
  }

    /**
     * converts a string to a float
     */
    private static float stod(String input) {
	return new Float(input).floatValue();
    }

    //number of required arguments needed on the command line
    private static short NUM_ARGUMENTS = 17;
    //number of optional arguments (used for framework generation)
    private static final short NUM_OPTIONAL_ARGUMENTS = 4;
  
    //number of arguments each distribution has
    private static final short SKEWED_NUM_PARAMS = 2;
    private static final short GAUSSIAN_NUM_PARAMS = 3;
    private static final short RANDOM_NUM_PARAMS = 1;
    
    private static int i = 0; //argument index

    static {
	//takes number of dimensions * 4 coordinates parameters at the end
	NUM_ARGUMENTS += NDIMS * 4;
    }
}
