package gstd;
/**
 * Generate 2D points or rectangles that are moving through time.
 * <p>
 * Implements the Generate_Spatio_Temporal_Data Algorithm as defined in [1]
 * and with the extensions described in [3].
 * <p>
 * Abstract class because outputInstance not defined.
 * <p>
 * Based on gstd.c
 *
 * @author <a href='mailto:mallett@cs.ualberta.ca'>Daniel Mallett</a>
 * @version %I%, %G%
 */
import java.io.*;
import java.util.*;

public abstract class GSTD_Abstract implements Constants {
    /**
     * All GSTD classes must implement output procs
     */
    public abstract void openOutput() throws IOException;
    public abstract void outputInstance(Instance instance) throws IOException;
    public abstract void closeOutput() throws IOException;
    
    public void setNumObjects(int numObjects) {
	intNumObjects = numObjects;
    }
    public void setStartingID(int startingID) {
	intStartingID = startingID;
    }
    public void setDensity(float density) {
	fltDensity = density;
    }
    public void setNumSnapShots(int numSnapShots) {
	intNumSnapShots = numSnapShots;
    }
    /**
     * create the Random number generator
     */
    public void setSeed(int seed) {
	intSeed = seed;
	random = new Random((long)seed);
    }
    public int getSeed() {
	return intSeed;
    }

    public void setObjectType(char objectType) {
	chrObjectType = objectType;
    }
    public void setApproach(char approach) {
	chrApproach = approach;
    }

    public void setInitialDistribution(Distribution distrib) {
	distribInitial = distrib;
    }
    public void setTimeDistribution(Distribution distrib) {
	distribTime = distrib;
    }
    public void setCenterDistribution(Distribution distrib) {
	distribCenter = distrib;
    }
    public void setExtentDistribution(Distribution distrib) {
	distribExtent = distrib;
    }
    public void setDIntervalDistribution(Distribution distrib) {
	distribDInterval = distrib;
    }
    public void setFramework(Framework framework) {
	this.framework = framework;
    }
    public Framework getFramework() {
	return framework;
    }

    /**
     * validate input
     */
    private boolean checkInput() {
	if (intNumObjects > MAXN) {
	    System.out.println("ERROR in #objects, maximum is " + MAXN);
	    return false;
	}
	return true;
    }

    public String getStartMsg() {
	return "Creating tuples for " + intNumObjects + " objects of:\n" +
	    "\tobject type:\t" + chrObjectType + "\n" +
	    "\tdistribution:\t" + 
	    distribInitial.getName() + "\n" +
	    "\tstore:\t\t" + chrApproach + "\n" +
	    "with now = " + NOW + " using seed = " + 
	    intSeed + "\n" +
	    "time interval follows distribution " + 
	    distribTime.getName();
    }
    private String getEndMsg(int counter) {
	return "... OK! generated " + counter + " tuples";
    }

    /**
     * Generates Spatio Temporal Data
     * assumes that all parameters have been set.
     *
     */
    public void generateSpatioTemporalData(boolean generateFramework,
					   boolean useFramework) {
	//validate input
	if (!checkInput())
	    return;

	//confirmation message
	System.out.println(getStartMsg());

	Instance[] arrCurrent = new Instance[intNumObjects];

	int counter = 0;
	int id = intStartingID;

	// open output
	try {
	    openOutput();
	} catch (IOException ioE) {
	    System.out.println("Error! - Unable to open output for writing.\n" + ioE.getMessage());
	    return;
	}

	// initialization
	for (int i = 0; i < intNumObjects; i++) {
	    // create a new object instance that is valid, 
	    // has a new space stamp, and a new timestamp
	    arrCurrent[i] = new Instance(id,true,
					 // initial space stamp is random center point in the unit workspace
					 // with initial extents that follow the density value
					 new SpaceStamp(chrObjectType,
							distribInitial.getRandom(random),
							getInitExtent(fltDensity / (float)intNumObjects)),
					 new TimeStamp());
	    
	    try {
		if (generateFramework) {
		    //keep generating until new rectangle doesn't intersect existing framework
		    //and is in the unit square
		    while (framework.intersects(arrCurrent[i].getSpaceStamp()) ||
			   !arrCurrent[i].getSpaceStamp().isInUnitSquare()) {
			arrCurrent[i] = new Instance(id,true, 
						     new SpaceStamp(chrObjectType,
								    distribInitial.getRandom(random),
								    getInitExtent(fltDensity / (float)intNumObjects)),
						     new TimeStamp());
		    }
		    //add rectangle to infrastructure framework
		    framework.addRectangle(arrCurrent[i].getSpaceStamp());

		} else if (useFramework) {
		    //using framework - make sure point does not fall inside infrastructure
		    framework.adjust(arrCurrent[i].getSpaceStamp(),null);
		}

		//normal initialization
		outputInstance(arrCurrent[i]);
		counter++;

	    } catch (IOException IOE) {
		System.out.println(IOE.getMessage());
	    }
	    id++;
	}

	if (intNumSnapShots > 1) {
	    // evolution
	    Instance instanceOld;
	    float fltInterval;
	    float fltIntervalSum = 0;
	    float fltDInterval = distribDInterval.getRandom(random)[0];
	    float fltStep = (float)NOW / (float)intNumSnapShots;
	    for (int i = 0; i < intNumObjects; i++) {
		boolean boolFlag = true;
		fltNextStep = fltStep;
		
		while (boolFlag) {
		    instanceOld = (Instance)arrCurrent[i].clone();
		    
		    // calculate new current values
		    do {
			fltInterval = distribTime.getRandom(random)[0];
		    } while (fltInterval <= 0);
		    fltIntervalSum += fltInterval;
		    
		    // handle dinterval
		    if (fltIntervalSum > fltDInterval) {
			fltDInterval = distribDInterval.getRandom(random)[0];
			
			// update delta center and delta extent (space stamp)
			arrCurrent[i].getSpaceStamp().update(distribCenter.getRandom(random),
							     distribExtent.getRandom(random));
			
			//check instance validity and adjust if necessary
			arrCurrent[i].setCurrent(arrCurrent[i].getSpaceStamp().adjust(chrApproach));		    
			if (useFramework) {
			    //using framework - adjust if necessary
			    framework.adjust(arrCurrent[i].getSpaceStamp(),instanceOld.getSpaceStamp());
			}

			// update timestamp
			arrCurrent[i].getTimeStamp().update(fltInterval);
			
			// reset interval
			fltIntervalSum = 0;
		    }
		    
		    // check instances and output
		    if (arrCurrent[i].getTimeStamp().getTransactionTime() > NOW) {
			boolFlag = false; // object is no longer active
			
			if (adjustTime(instanceOld.getTimeStamp(),
				       arrCurrent[i].getTimeStamp().getTransactionTime(),
				       fltStep)) {
			    try {
				outputInstance(instanceOld);
				counter++;
			    } catch (IOException IOE) {
				System.out.println(IOE.getMessage());
			    }
			}
		    } else {
			if (arrCurrent[i].getTimeStamp().getTransactionTime() > fltNextStep) {
			    if (adjustTime(instanceOld.getTimeStamp(),
					   arrCurrent[i].getTimeStamp().getTransactionTime(),
					   fltStep)) {
				try {
				    outputInstance(instanceOld);
				    counter++;
				} catch (IOException IOE) {
				    System.out.println(IOE.getMessage());
				}
			    }
			}
		    }
		}
	    }
	}
	// close output
	try {
	    closeOutput();
	} catch (IOException ioE) {
	    System.out.println("Error! - Unable to close output.\n" + ioE.getMessage());
	    return;
	}
	// end message
	System.out.println(getEndMsg(counter));
    }

    /**
     * adjust the instance time
     *
     */
    private boolean adjustTime(TimeStamp ttOld, float ttCurrent, float step) {
	if (ttOld.getTransactionTime() == 0) {
	    fltNextStep = (int)(ttOld.getTransactionTime() * intNumSnapShots) * step + step;
	    // old instance has already been printed
	    return false;
	} else if ( (int)(ttCurrent * intNumSnapShots) > (int)((fltNextStep * intNumSnapShots) + ERR)) {
	    // test if old instance is exactly a snapshot time step
	    if ( ((ttOld.getTransactionTime() * intNumSnapShots) - 
		  (int)(ttOld.getTransactionTime() * intNumSnapShots)) > 0)
		//it is not
		ttOld.setTransactionTime((int)(ttOld.getTransactionTime() * intNumSnapShots) * step + step);
	    else
		// it is
		ttOld.setTransactionTime((int)(ttOld.getTransactionTime() * intNumSnapShots) * step);

	    ttOld.setTransactionTime(Math.min(ttOld.getTransactionTime(),1));

	    if (ttCurrent > ttOld.getTransactionTime())
		fltNextStep = ((int)(ttCurrent * intNumSnapShots) * step) + step;
	    else
		fltNextStep = ttOld.getTransactionTime() + step;
	} else {
	    if ((int)(ttOld.getTransactionTime() * intNumSnapShots) == (int)(ttCurrent * intNumSnapShots))
		return false;
	    else {
		ttOld.setTransactionTime((int)(ttOld.getTransactionTime() * intNumSnapShots) * step + step);
		fltNextStep = ttOld.getTransactionTime() + step;
	    }
	}
	return true;
    }


    /**
     * returns the initial extents for a rectangle (follows density)
     * extent[] = extent(N,D)
     *
     * @param area is the density/number objects
     *
     */
    private float[] getInitExtent(float area) {
	//angle = PI * uniform(0,90) / 180
	//x = sqrt(area * tan(angle))
	float number = new DistributionRandom(new float[] {(float)0.0},
					      new float[] {(float)90.0}).getRandom(random)[0];
	float angle = (float)((Math.PI * number) / 180.0);
	
	float tan = (float)Math.tan(angle);

	float sqrt = (float)Math.sqrt(area*tan);

	//adjust in case angle too close to 90 gives bad extent value
	if (sqrt > 1) {
	    sqrt = sqrt - 1;
	}
	float[] retval = new float[NDIMS];
	
	/* XXX
	 * DOES NOT handle handle dimensions higher than 2 properly
	 * XXX
	 */
	retval[0]  = (area/sqrt);
	for (int i = 1; i < NDIMS; i++)
	    retval[i] = sqrt;
	
	return retval;
    }

    //random number generator
    private Random random;
    private float fltNextStep = 0;

    //GSTD parameters - protected because output class might want to use these
    protected int intNumObjects = -1;   //cardinality
    protected int intStartingID = -1;   //first id of the objects
    protected float fltDensity = (float)-1.0;   //density of the objects
    protected int intNumSnapShots = -1; //number of snapshots
    protected int intSeed = -1;         //seed of random function
    protected char chrObjectType;  //type of objects: points or rectangles
    protected char chrApproach;    //adjument or radar approach

    //GSTD distribution parmaters
    protected Distribution distribInitial;   // initial distribution
    protected Distribution distribTime;      // distribution of the interval time
    protected Distribution distribCenter;    // distribution of the center
    protected Distribution distribExtent;    // distribution of rectangle extent
    protected Distribution distribDInterval; // distribution of dinterval

    //framework
    protected Framework framework; //if infrastructure present, framework != null

    //output file
    protected String strFileName;
}
